from typing import ClassVar, Dict, List, Tuple, Type, overload

import numpy as np
import numpy.typing as npt

class NodeStatus:
    __members__: ClassVar[dict] = ...  # read-only
    CORE: ClassVar[NodeStatus] = ...
    FIXED_VALUE: ClassVar[NodeStatus] = ...
    FIXED_GRADIENT: ClassVar[NodeStatus] = ...
    LOOPED: ClassVar[NodeStatus] = ...
    __entries: ClassVar[dict] = ...
    def __init__(self, value: int) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    def __setstate__(self, state: int) -> None: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class Node:
    def __init__(self, idx: int, status: NodeStatus) -> None: ...
    idx: int
    status: NodeStatus

class Neighbor:
    def __init__(self, idx: int, distance: float, status: NodeStatus) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    idx: int
    distance: float
    status: NodeStatus

class ProfileBoundaryStatus:
    @overload
    def __init__(self, status: NodeStatus) -> None: ...
    @overload
    def __init__(self, left_status: NodeStatus, right_status: NodeStatus) -> None: ...
    @overload
    def __init__(self, status: List[NodeStatus]) -> None: ...
    left: NodeStatus
    right: NodeStatus
    @property
    def is_horizontal_looped(self) -> bool: ...

class ProfileGrid:
    def __init__(
        self,
        size: int,
        spacing: float,
        bounds_status: NodeStatus | List[NodeStatus] | ProfileBoundaryStatus,
        nodes_status: Dict[int, NodeStatus] | None = None,
    ) -> None: ...
    @classmethod
    def from_length(
        cls: Type[ProfileGrid],
        size: int,
        length: float,
        bounds_status: NodeStatus | List[NodeStatus] | ProfileBoundaryStatus,
        nodes_status: Dict[int, NodeStatus] | None = None,
    ) -> ProfileGrid: ...
    is_structured: ClassVar[bool]
    is_uniform: ClassVar[bool]
    n_neighbors_max: ClassVar[int]
    @property
    def size(self) -> int: ...
    @property
    def shape(self) -> List[int]: ...
    @property
    def spacing(self) -> float: ...
    @property
    def length(self) -> float: ...
    @overload
    def nodes_indices(self) -> npt.NDArray[np.uint64]: ...
    @overload
    def nodes_indices(self, status: NodeStatus) -> npt.NDArray[np.uint64]: ...
    @overload
    def nodes_status(self) -> npt.NDArray[np.uint8]: ...
    @overload
    def nodes_status(self, idx: int) -> NodeStatus: ...
    @overload
    def nodes_areas(self) -> npt.NDArray[np.double]: ...
    @overload
    def nodes_areas(self, idx: int) -> float: ...
    def neighbors_count(self, idx: int) -> int: ...
    def neighbors_indices(self, idx: int) -> npt.NDArray[np.uint64]: ...
    def neighbors_distances(self, idx: int) -> npt.NDArray[np.float64]: ...
    def neighbors(self, idx: int) -> List[Neighbor]: ...

class RasterNode:
    def __init__(self, row: int, col: int, status: NodeStatus) -> None: ...
    row: int
    col: int
    status: NodeStatus

class RasterNeighbor:
    def __init__(
        self, flatten_idx: int, row: int, col: int, distance: float, status: NodeStatus
    ) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    flatten_idx: int
    row: int
    col: int
    distance: float
    status: NodeStatus

class RasterBoundaryStatus:
    @overload
    def __init__(self, status: NodeStatus) -> None: ...
    @overload
    def __init__(self, status: List[NodeStatus]) -> None: ...
    left: NodeStatus
    right: NodeStatus
    bottom: NodeStatus
    top: NodeStatus
    @property
    def is_horizontal_looped(self) -> bool: ...
    @property
    def is_vertical_looped(self) -> bool: ...

class RasterGrid:
    def __init__(
        self,
        shape: List[int],
        spacing: npt.ArrayLike,
        bounds_status: NodeStatus | List[NodeStatus] | RasterBoundaryStatus,
        nodes_status: Dict[Tuple[int, int], NodeStatus] | None = None,
    ) -> None: ...
    @classmethod
    def from_length(
        cls: Type[RasterGrid],
        shape: List[int],
        length: npt.ArrayLike,
        bounds_status: NodeStatus | List[NodeStatus] | RasterBoundaryStatus,
        nodes_status: Dict[Tuple[int, int], NodeStatus] | None = None,
    ) -> RasterGrid: ...
    is_structured: ClassVar[bool]
    is_uniform: ClassVar[bool]
    n_neighbors_max: ClassVar[int]
    @property
    def size(self) -> int: ...
    @property
    def shape(self) -> List[int]: ...
    @property
    def spacing(self) -> float: ...
    @property
    def length(self) -> float: ...
    @overload
    def nodes_indices(self) -> npt.NDArray[np.uint64]: ...
    @overload
    def nodes_indices(self, status: NodeStatus) -> npt.NDArray[np.uint64]: ...
    @overload
    def nodes_status(self) -> npt.NDArray[np.uint8]: ...
    @overload
    def nodes_status(self, idx: int) -> NodeStatus: ...
    @overload
    def nodes_areas(self) -> npt.NDArray[np.double]: ...
    @overload
    def nodes_areas(self, idx: int) -> float: ...
    def neighbors_count(self, idx: int) -> int: ...
    @overload
    def neighbors_indices(self, idx: int) -> npt.NDArray[np.uint64]: ...
    @overload
    def neighbors_indices(self, row: int, col: int) -> npt.NDArray[np.uint64]: ...
    def neighbors_distances(self, idx: int) -> npt.NDArray[np.float64]: ...
    @overload
    def neighbors(self, idx: int) -> List[Neighbor]: ...
    @overload
    def neighbors(self, row: int, col: int) -> List[RasterNeighbor]: ...

class TriMesh:
    @overload
    def __init__(
        self,
        points: npt.NDArray[np.float64],
        triangles: npt.NDArray[np.uint64],
        nodes_status: Dict[int, NodeStatus] | None = None,
    ) -> None: ...
    @overload
    def __init__(
        self,
        points: npt.NDArray[np.float64],
        triangles: npt.NDArray[np.uint64],
        nodes_status: npt.ArrayLike,
    ) -> None: ...
    is_structured: ClassVar[bool]
    is_uniform: ClassVar[bool]
    n_neighbors_max: ClassVar[int]
    @property
    def size(self) -> int: ...
    @property
    def shape(self) -> List[int]: ...
    @overload
    def nodes_indices(self) -> npt.NDArray[np.uint64]: ...
    @overload
    def nodes_indices(self, status: NodeStatus) -> npt.NDArray[np.uint64]: ...
    @overload
    def nodes_status(self) -> npt.NDArray[np.uint8]: ...
    @overload
    def nodes_status(self, idx: int) -> NodeStatus: ...
    @overload
    def nodes_areas(self) -> npt.NDArray[np.double]: ...
    @overload
    def nodes_areas(self, idx: int) -> float: ...
    def neighbors_count(self, idx: int) -> int: ...
    def neighbors_indices(self, idx: int) -> npt.NDArray[np.uint64]: ...
    def neighbors_distances(self, idx: int) -> npt.NDArray[np.float64]: ...
    def neighbors(self, idx: int) -> List[Neighbor]: ...



from typing import ClassVar, List, Union, overload

import numpy as np
import numpy.typing as npt

from fastscapelib.grid import ProfileGrid, RasterGrid, TriMesh

Grid = Union[ProfileGrid, RasterGrid, TriMesh]

class FlowGraphImpl:
    @property
    def single_flow(self) -> bool: ...
    @property
    def receivers(self) -> npt.NDArray[np.uint64]: ...
    @property
    def receivers_count(self) -> npt.NDArray[np.uint64]: ...
    @property
    def receivers_distance(self) -> npt.NDArray[np.float64]: ...
    @property
    def receivers_weight(self) -> npt.NDArray[np.float64]: ...
    @property
    def donors(self) -> npt.NDArray[np.uint64]: ...
    @property
    def donors_count(self) -> npt.NDArray[np.uint64]: ...
    @property
    def dfs_indices(self) -> npt.NDArray[np.uint64]: ...
    @property
    def basins(self) -> npt.NDArray[np.uint64]: ...

class FlowDirection:
    __members__: ClassVar[dict] = ...  # read-only
    UNDEFINED: ClassVar[FlowDirection] = ...
    SINGLE: ClassVar[FlowDirection] = ...
    MULTI: ClassVar[FlowDirection] = ...
    __entries: ClassVar[dict] = ...
    def __init__(self, value: int) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    def __setstate__(self, state: int) -> None: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class FlowOperator:
    @property
    def name(self) -> str: ...

class SingleFlowRouter(FlowOperator):
    def __init__(self) -> None: ...
    graph_updated: ClassVar[bool] = ...
    elevation_updated: ClassVar[bool] = ...
    in_flowdir: ClassVar[FlowDirection] = ...
    out_flowdir: ClassVar[FlowDirection] = ...
    def __repr__(self) -> str: ...

class MultiFlowRouter(FlowOperator):
    def __init__(self, slope_exp: float = 1.0) -> None: ...
    graph_updated: ClassVar[bool] = ...
    elevation_updated: ClassVar[bool] = ...
    in_flowdir: ClassVar[FlowDirection] = ...
    out_flowdir: ClassVar[FlowDirection] = ...
    slope_exp: float
    def __repr__(self) -> str: ...

class PFloodSinkResolver(FlowOperator):
    def __init__(self) -> None: ...
    graph_updated: ClassVar[bool] = ...
    elevation_updated: ClassVar[bool] = ...
    in_flowdir: ClassVar[FlowDirection] = ...
    out_flowdir: ClassVar[FlowDirection] = ...
    def __repr__(self) -> str: ...

class MSTMethod:
    __members__: ClassVar[dict] = ...  # read-only
    KRUSKAL: ClassVar[MSTMethod] = ...
    BORUVKA: ClassVar[MSTMethod] = ...
    __entries: ClassVar[dict] = ...
    def __init__(self, value: int) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    def __setstate__(self, state: int) -> None: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class MSTRouteMethod:
    __members__: ClassVar[dict] = ...  # read-only
    BASIC: ClassVar[MSTRouteMethod] = ...
    CARVE: ClassVar[MSTRouteMethod] = ...
    __entries: ClassVar[dict] = ...
    def __init__(self, value: int) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    def __setstate__(self, state: int) -> None: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class MSTSinkResolver(FlowOperator):
    def __init__(
        self,
        basin_method: MSTMethod = MSTMethod.KRUSKAL,
        route_method: MSTRouteMethod = MSTRouteMethod.CARVE,
    ) -> None: ...
    graph_updated: ClassVar[bool] = ...
    elevation_updated: ClassVar[bool] = ...
    in_flowdir: ClassVar[FlowDirection] = ...
    out_flowdir: ClassVar[FlowDirection] = ...
    basin_method: MSTMethod
    route_method: MSTRouteMethod
    def __repr__(self) -> str: ...

class FlowSnapshot(FlowOperator):
    def __init__(
        self, snapshot_name: str, save_graph: bool = True, save_elevation: bool = False
    ) -> None: ...
    graph_updated: ClassVar[bool] = ...
    elevation_updated: ClassVar[bool] = ...
    in_flowdir: ClassVar[FlowDirection] = ...
    out_flowdir: ClassVar[FlowDirection] = ...
    @property
    def snapshot_name(self) -> str: ...
    @property
    def save_graph(self) -> bool: ...
    @property
    def save_elevation(self) -> bool: ...
    def __repr__(self) -> str: ...

class FlowGraph:
    def __init__(self, grid: Grid, operators: List[FlowOperator]) -> None: ...
    @property
    def operators(self) -> List[FlowOperator]: ...
    @property
    def single_flow(self) -> bool: ...
    def impl(self) -> FlowGraphImpl: ...
    @property
    def graph_snapshot_keys(self) -> List[str]: ...
    def graph_snapshot(self, name: str) -> FlowGraph: ...
    @property
    def elevation_snapshot_keys(self) -> List[str]: ...
    def elevation_snapshot(self, name: str) -> npt.NDArray[np.float64]: ...
    def update_routes(
        self, elevation: npt.NDArray[np.float64]
    ) -> npt.NDArray[np.float64]: ...
    @property
    def base_levels(self) -> list[int]: ...
    @base_levels.setter
    def base_levels(self, value: list[int]) -> None: ...
    @property
    def mask(self) -> npt.NDArray[np.bool_]: ...
    @mask.setter
    def mask(self, value: npt.NDArray[np.bool_]) -> None: ...
    @overload
    def accumulate(
        self, acc: npt.NDArray[np.float64], src: npt.NDArray[np.float64]
    ) -> None: ...
    @overload
    def accumulate(self, acc: npt.NDArray[np.float64], src: float) -> None: ...
    @overload
    def accumulate(self, src: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]: ...
    @overload
    def accumulate(self, src: float) -> npt.NDArray[np.float64]: ...
    def basins(self) -> npt.NDArray[np.uint64]: ...
    def __repr__(self) -> str: ...



from _fastscapelib_py.grid import (  # type: ignore[import]
    Neighbor,
    Node,
    NodeStatus,
    ProfileBoundaryStatus,
    ProfileGrid,
    RasterBoundaryStatus,
    RasterGrid,
    RasterNeighbor,
    RasterNode,
    TriMesh,
)

__all__ = [
    "Neighbor",
    "Node",
    "NodeStatus",
    "ProfileBoundaryStatus",
    "ProfileGrid",
    "RasterBoundaryStatus",
    "RasterGrid",
    "RasterNeighbor",
    "RasterNode",
    "TriMesh",
]



import numpy as np
import numpy.testing as npt
import pytest

from fastscapelib.grid import Neighbor, NodeStatus, TriMesh


@pytest.fixture
def mesh_args():
    # simple mesh with four boundary nodes and one inner node:
    #      o
    #   o  o  o
    #      o
    #
    # Returns arguments passed to the TriMesh constructor

    points = np.array([[0.0, 0.5], [0.5, 0.0], [0.0, -0.5], [-0.5, 0.0], [0.0, 0.0]])
    triangles = np.array([[2, 4, 3], [4, 2, 1], [4, 0, 3], [0, 4, 1]])

    return points, triangles


class TestTriMesh:
    def test_static_properties(self) -> None:
        assert TriMesh.is_structured is False
        assert TriMesh.is_uniform is False
        assert TriMesh.n_neighbors_max == 20

    def test_constructor(self, mesh_args) -> None:
        points, triangles = mesh_args
        mesh = TriMesh(points, triangles)

        assert mesh.size == 5
        assert mesh.shape == [5]

        with pytest.raises(ValueError, match="invalid shape for points array"):
            TriMesh(np.array([[0.0, 0.5, 1.0]]), triangles)

        with pytest.raises(ValueError, match="invalid shape for triangles array"):
            TriMesh(points, np.array([[0, 1]]))

        with pytest.raises(ValueError, match="invalid shape for nodes_status array"):
            TriMesh(points, triangles, [0])

    def test_nodes_indices(self, mesh_args) -> None:
        mesh = TriMesh(*mesh_args)
        npt.assert_equal(mesh.nodes_indices(), np.arange(mesh.size))
        npt.assert_equal(mesh.nodes_indices(NodeStatus.FIXED_VALUE), [0, 1, 2, 3])
        npt.assert_equal(mesh.nodes_indices(NodeStatus.CORE), [4])
        assert not len(mesh.nodes_indices(NodeStatus.FIXED_GRADIENT))

    def test_nodes_status_default(self, mesh_args) -> None:
        points, triangles = mesh_args

        # all boundary nodes have fixed value status
        mesh = TriMesh(points, triangles)

        actual = mesh.nodes_status()
        expected = [1, 1, 1, 1, 0]

        npt.assert_array_equal(actual, expected)

        # should return a copy
        assert not np.shares_memory(mesh.nodes_status(), mesh.nodes_status())

        assert mesh.nodes_status(0) == NodeStatus.FIXED_VALUE
        assert mesh.nodes_status(4) == NodeStatus.CORE

        # initialize with an array of node status
        mesh2 = TriMesh(points, triangles, [1, 1, 0, 1, 1])
        npt.assert_array_equal(mesh2.nodes_status(), [1, 1, 0, 1, 1])

    def test_nodes_status_custom(self, mesh_args) -> None:
        points, triangles = mesh_args
        mesh = TriMesh(points, triangles, {2: NodeStatus.FIXED_VALUE})

        actual = mesh.nodes_status()
        expected = np.zeros(mesh.size, dtype=np.uint8)
        expected[2] = 1

        npt.assert_array_equal(actual, expected)

        with pytest.raises(ValueError, match=".*not allowed.*"):
            TriMesh(points, triangles, {2: NodeStatus.LOOPED})

    def test_nodes_areas(self) -> None:
        # simple case of 1x1 domain with 9 evenly spaced nodes
        points = np.array(
            [
                [0, 0],
                [0, 0.5],
                [0, 1],
                [0.5, 1],
                [1, 1],
                [1, 0.5],
                [1, 0],
                [0.5, 0],
                [0.5, 0.5],
            ]
        )
        triangles = np.array(
            [
                [0, 1, 8],
                [1, 2, 8],
                [2, 3, 8],
                [3, 4, 8],
                [4, 5, 8],
                [5, 6, 8],
                [6, 7, 8],
                [7, 0, 8],
            ]
        )
        mesh = TriMesh(points, triangles)

        assert mesh.nodes_areas(0) == 0.0625
        assert mesh.nodes_areas(8) == 0.25
        npt.assert_equal(
            mesh.nodes_areas(),
            [0.0625, 0.125, 0.0625, 0.125, 0.0625, 0.125, 0.0625, 0.125, 0.25],
        )

    def test_neighbors_count(self, mesh_args) -> None:
        mesh = TriMesh(*mesh_args)

        assert mesh.neighbors_count(4) == 4
        assert mesh.neighbors_count(0) == 3

    def test_neighbors_indices(self, mesh_args) -> None:
        mesh = TriMesh(*mesh_args)

        npt.assert_equal(np.sort(mesh.neighbors_indices(3)), [0, 2, 4])
        npt.assert_equal(np.sort(mesh.neighbors_indices(4)), [0, 1, 2, 3])

    def test_neighbors_distances(self, mesh_args) -> None:
        mesh = TriMesh(*mesh_args)

        dist_diag = np.sqrt(0.5**2 + 0.5**2)

        sorted_nidx = np.argsort(mesh.neighbors_indices(3)).ravel()
        npt.assert_equal(
            mesh.neighbors_distances(3)[sorted_nidx], [dist_diag, dist_diag, 0.5]
        )

        npt.assert_equal(mesh.neighbors_distances(4), [0.5] * 4)

    def test_neighbors(self, mesh_args) -> None:
        mesh = TriMesh(*mesh_args)

        dist_diag = np.sqrt(0.5**2 + 0.5**2)

        sorted_nidx = np.argsort(mesh.neighbors_indices(3)).ravel()
        temp = mesh.neighbors(3)
        actual = [temp[i] for i in sorted_nidx]

        expected = [
            Neighbor(0, dist_diag, NodeStatus.FIXED_VALUE),
            Neighbor(2, dist_diag, NodeStatus.FIXED_VALUE),
            Neighbor(4, 0.5, NodeStatus.CORE),
        ]

        assert actual == expected

        with pytest.raises(IndexError, match="grid index out of range"):
            mesh.neighbors(111)



import numpy as np
import numpy.testing as npt
import pytest

from fastscapelib.flow import (
    FlowDirection,
    FlowGraph,
    FlowOperator,
    FlowSnapshot,
    MSTSinkResolver,
    PFloodSinkResolver,
    SingleFlowRouter,
)
from fastscapelib.grid import NodeStatus, RasterGrid


@pytest.fixture
def grid() -> RasterGrid:
    return RasterGrid([10, 20], [1.0, 2.0], NodeStatus.FIXED_VALUE)


def test_flow_snapshot_class_attr() -> None:
    assert FlowSnapshot.graph_updated is False
    assert FlowSnapshot.elevation_updated is False
    assert FlowSnapshot.in_flowdir == FlowDirection.UNDEFINED
    assert FlowSnapshot.out_flowdir == FlowDirection.UNDEFINED


def test_flow_snapshot_constructor() -> None:
    snapshot = FlowSnapshot("test")
    assert isinstance(snapshot, FlowOperator)
    assert snapshot.name == "flow_snapshot"
    assert snapshot.snapshot_name == "test"
    assert snapshot.save_graph is True
    assert snapshot.save_elevation is False

    assert repr(snapshot) == "FlowSnapshot 'test' (graph=True, elevation=False)"


def test_flow_snapshot_error(grid) -> None:
    with pytest.raises(ValueError, match="no flow routing operator defined before"):
        FlowGraph(grid, [FlowSnapshot("test"), SingleFlowRouter()])


def test_flow_snapshot_items(grid) -> None:
    graph = FlowGraph(
        grid,
        [
            PFloodSinkResolver(),
            FlowSnapshot("a", save_graph=False, save_elevation=True),
            SingleFlowRouter(),
            FlowSnapshot("b", save_graph=True, save_elevation=False),
        ],
    )

    assert graph.graph_snapshot_keys == ["b"]
    assert graph.elevation_snapshot_keys == ["a"]

    assert isinstance(graph.graph_snapshot("b"), FlowGraph)
    assert isinstance(graph.elevation_snapshot("a"), np.ndarray)


def test_snapshot_graph(grid) -> None:
    graph = FlowGraph(
        grid,
        [SingleFlowRouter(), FlowSnapshot("a"), MSTSinkResolver(), FlowSnapshot("b")],
    )

    elevation = np.random.uniform(size=grid.shape)
    graph.update_routes(elevation)
    snapshot_a = graph.graph_snapshot("a")
    snapshot_b = graph.graph_snapshot("b")

    attrs = [
        "receivers_count",
        "receivers",
        "receivers_distance",
        "receivers_weight",
        "donors_count",
        "dfs_indices",
    ]

    for name in attrs:
        arr = getattr(graph.impl(), name)
        snapshot_a_arr = getattr(snapshot_a.impl(), name)
        snapshot_b_arr = getattr(snapshot_b.impl(), name)

        assert arr is not snapshot_a_arr

        npt.assert_equal(arr, snapshot_b_arr)
        assert arr is not snapshot_b_arr

    npt.assert_equal(snapshot_b.impl().donors[:, 0], graph.impl().donors[:, 0])
    assert not np.all(snapshot_a.impl().receivers == snapshot_b.impl().receivers)

    # high-level interface test
    actual = graph.accumulate(1.0)
    expected = snapshot_b.accumulate(1.0)
    npt.assert_allclose(actual, expected)

    # test snapshot graphs are read-only
    with pytest.raises(RuntimeError, match=".*(read-only)"):
        snapshot_a.update_routes(elevation)


def test_snapshot_graph_impl(grid) -> None:
    # Ensure that this will not crash Python (segfault). The correct pybind11
    # return value policy for accessing the implementation from a snapshot graph
    # is "reference_internal" (we need the keep_alive call policy to prevent the
    # parent Python graph / snapshot / impl wrappers getting garbage collected
    # when accessing the final graph implementation data members).
    graph = FlowGraph(
        grid,
        [SingleFlowRouter(), FlowSnapshot("a"), MSTSinkResolver(), FlowSnapshot("b")],
    )

    elevation = np.random.uniform(size=grid.shape)
    graph.update_routes(elevation)

    graph.graph_snapshot("a").impl().receivers
    s = graph.graph_snapshot("a")
    s.impl().receivers
    simpl = s.impl()
    simpl.receivers


def test_snapshot_elevation(grid) -> None:
    graph = FlowGraph(
        grid,
        [
            FlowSnapshot("a", save_graph=False, save_elevation=True),
            PFloodSinkResolver(),
            FlowSnapshot("b", save_graph=False, save_elevation=True),
            SingleFlowRouter(),
        ],
    )

    elevation = np.random.uniform(size=grid.shape)
    filled_elevation = graph.update_routes(elevation)
    snapshot_a = graph.elevation_snapshot("a")
    snapshot_b = graph.elevation_snapshot("b")

    npt.assert_equal(elevation, snapshot_a)
    assert elevation is not snapshot_a
    npt.assert_equal(filled_elevation, snapshot_b)
    assert filled_elevation is not snapshot_b



import numpy as np
import numpy.testing as npt
import pytest

from fastscapelib.flow import (
    FlowDirection,
    FlowGraph,
    FlowOperator,
    MultiFlowRouter,
    PFloodSinkResolver,
    SingleFlowRouter,
)
from fastscapelib.grid import NodeStatus, ProfileGrid, RasterGrid


class TestSingleFlowRouter:
    profile_flow_graph: FlowGraph
    profile_elevation: np.ndarray
    result_profile_elevation: np.ndarray
    raster_flow_graph: FlowGraph
    raster_elevation: np.ndarray
    result_raster_elevation: np.ndarray

    @classmethod
    def setup_class(cls) -> None:
        profile_grid = ProfileGrid(8, 2.2, NodeStatus.FIXED_VALUE)
        cls.profile_flow_graph = FlowGraph(profile_grid, [SingleFlowRouter()])
        cls.profile_elevation = np.array([0.0, 0.2, 0.1, 0.2, 0.4, 0.6, 0.3, 0.0])
        cls.result_profile_elevation = cls.profile_flow_graph.update_routes(
            cls.profile_elevation
        )

        # bottom border base level
        bottom_base_level = [
            NodeStatus.CORE,
            NodeStatus.CORE,
            NodeStatus.CORE,
            NodeStatus.FIXED_VALUE,
        ]

        raster_grid = RasterGrid([4, 4], [1.0, 1.0], bottom_base_level)

        cls.raster_flow_graph = FlowGraph(raster_grid, [SingleFlowRouter()])

        # planar surface tilted along the y-axis + small carved channel
        cls.raster_elevation = np.array(
            [
                [0.6, 0.6, 0.6, 0.6],
                [0.4, 0.4, 0.4, 0.4],
                [0.2, 0.2, 0.2, 0.2],
                [0.1, 0.0, 0.1, 0.1],
            ]
        )
        cls.result_raster_elevation = cls.raster_flow_graph.update_routes(
            cls.raster_elevation
        )

    def test_constructor(self) -> None:
        router = SingleFlowRouter()
        assert isinstance(router, FlowOperator)
        assert router.name == "single_flow_router"
        assert repr(router) == "SingleFlowRouter"

    def test_class_attrs(self) -> None:
        assert SingleFlowRouter.graph_updated is True
        assert SingleFlowRouter.elevation_updated is False
        assert SingleFlowRouter.in_flowdir == FlowDirection.UNDEFINED
        assert SingleFlowRouter.out_flowdir == FlowDirection.SINGLE

    def test_receivers(self) -> None:
        npt.assert_equal(
            self.profile_flow_graph.impl().receivers[:, 0],
            np.array([0, 0, 2, 2, 3, 6, 7, 7]),
        )

        npt.assert_equal(
            self.raster_flow_graph.impl().receivers[:, 0],
            np.array([4, 5, 6, 7, 8, 9, 10, 11, 13, 13, 13, 15, 12, 13, 14, 15]),
        )

    def test_receivers_count(self) -> None:
        npt.assert_equal(self.profile_flow_graph.impl().receivers_count, np.ones(8))

        npt.assert_equal(self.raster_flow_graph.impl().receivers_count, np.ones(16))

    def test_receivers_distance(self) -> None:
        npt.assert_equal(
            self.profile_flow_graph.impl().receivers_distance[:, 0],
            np.array([0, 1, 0, 1, 1, 1, 1, 0]) * 2.2,
        )

        dia = np.sqrt(2)
        npt.assert_equal(
            self.raster_flow_graph.impl().receivers_distance[:, 0],
            np.array(
                [
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    1.0,
                    dia,
                    1.0,
                    dia,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                ]
            ),
        )

    def test_receivers_weight(self) -> None:
        npt.assert_equal(
            self.profile_flow_graph.impl().receivers_weight[:, 0], np.ones(8)
        )
        npt.assert_equal(
            self.raster_flow_graph.impl().receivers_weight[:, 0], np.ones(16)
        )

    def test_donors(self) -> None:
        actual = self.profile_flow_graph.impl().donors
        m = np.iinfo(actual.dtype).max
        expected = np.array(
            [
                [1, m, m],
                [m, m, m],
                [2, 3, m],
                [4, m, m],
                [m, m, m],
                [m, m, m],
                [5, m, m],
                [6, m, m],
            ]
        )
        npt.assert_equal(actual, expected)

        actual = self.raster_flow_graph.impl().donors
        m = np.iinfo(actual.dtype).max
        expected = np.array(
            [
                [m, m, m, m, m, m, m, m, m],
                [m, m, m, m, m, m, m, m, m],
                [m, m, m, m, m, m, m, m, m],
                [m, m, m, m, m, m, m, m, m],
                [0, m, m, m, m, m, m, m, m],
                [1, m, m, m, m, m, m, m, m],
                [2, m, m, m, m, m, m, m, m],
                [3, m, m, m, m, m, m, m, m],
                [4, m, m, m, m, m, m, m, m],
                [5, m, m, m, m, m, m, m, m],
                [6, m, m, m, m, m, m, m, m],
                [7, m, m, m, m, m, m, m, m],
                [m, m, m, m, m, m, m, m, m],
                [8, 9, 10, m, m, m, m, m, m],
                [m, m, m, m, m, m, m, m, m],
                [11, m, m, m, m, m, m, m, m],
            ]
        )
        npt.assert_equal(actual, expected)

    def test_donors_count(self) -> None:
        npt.assert_equal(
            self.profile_flow_graph.impl().donors_count,
            np.array([1, 0, 2, 1, 0, 0, 1, 1]),
        )

        npt.assert_equal(
            self.raster_flow_graph.impl().donors_count,
            np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 3, 0, 1]),
        )

    def test_dfs_indices(self) -> None:
        npt.assert_equal(
            self.profile_flow_graph.impl().dfs_indices,
            np.array([0, 1, 2, 3, 4, 7, 6, 5]),
        )

        npt.assert_equal(
            self.raster_flow_graph.impl().dfs_indices,
            np.array([12, 13, 8, 9, 10, 6, 2, 5, 1, 4, 0, 14, 15, 11, 7, 3]),
        )


class TestMultiFlowRouter:
    def test_constructor(self) -> None:
        router = MultiFlowRouter()
        assert isinstance(router, FlowOperator)
        assert router.name == "multi_flow_router"
        assert repr(router).startswith("MultiFlowRouter (slope_exp=1.0")
        assert router.slope_exp == 1.0

        router = MultiFlowRouter(0.0)
        assert router.slope_exp == 0.0

        router.slope_exp = 0.5
        assert router.slope_exp == 0.5

    def test_class_attrs(self) -> None:
        assert MultiFlowRouter.graph_updated is True
        assert MultiFlowRouter.elevation_updated is False
        assert MultiFlowRouter.in_flowdir == FlowDirection.UNDEFINED
        assert MultiFlowRouter.out_flowdir == FlowDirection.MULTI

    def test_graph_topology(self) -> None:
        # test on a 3x3 tiny grid with base levels at all border nodes
        grid = RasterGrid([3, 3], [1.0, 1.0], NodeStatus.FIXED_VALUE)

        diag = np.sqrt(2)
        elevation = np.array([[0.0, 1.0, 0.0], [1.0, 2.0, 1.0], [0.0, 1.0, 0.0]])
        router = MultiFlowRouter(0.0)
        flow_graph = FlowGraph(grid, [router])
        flow_graph.update_routes(elevation)

        npt.assert_array_equal(
            flow_graph.impl().receivers_count,
            [1, 1, 1, 1, 8, 1, 1, 1, 1],
        )
        npt.assert_array_equal(
            flow_graph.impl().receivers[4],
            [0, 1, 2, 3, 5, 6, 7, 8],
        )
        npt.assert_array_almost_equal(
            flow_graph.impl().receivers_distance[4],
            [diag, 1.0, diag, 1.0, 1.0, diag, 1.0, diag],
        )
        npt.assert_array_equal(
            flow_graph.impl().donors_count,
            [1, 1, 1, 1, 0, 1, 1, 1, 1],
        )
        npt.assert_array_equal(
            flow_graph.impl().donors[[0, 1, 2, 3, 5, 6, 7, 8], 0].ravel(),
            [4, 4, 4, 4, 4, 4, 4, 4],
        )

    def test_mask(self) -> None:
        # see test_graph_topology for test setup
        grid = RasterGrid([3, 3], [1.0, 1.0], NodeStatus.FIXED_VALUE)

        elevation = np.array([[0.0, 1.0, 0.0], [1.0, 2.0, 1.0], [0.0, 1.0, 0.0]])

        router = MultiFlowRouter(0.0)
        flow_graph = FlowGraph(grid, [router])

        # masked center node -> no graph edge (each node has itself as receiver)
        mask = np.array(
            [[False, False, False], [False, True, False], [False, False, False]]
        )
        flow_graph.mask = mask
        flow_graph.update_routes(elevation)
        npt.assert_array_equal(flow_graph.impl().receivers_count, np.ones(9, dtype=int))

        # one masked edge node -> this masked node not in center node receivers
        mask = np.array(
            [[False, False, False], [True, False, False], [False, False, False]]
        )
        flow_graph.mask = mask
        flow_graph.update_routes(elevation)
        npt.assert_array_equal(
            flow_graph.impl().receivers[4][:-1],
            [0, 1, 2, 5, 6, 7, 8],
        )

    @pytest.mark.parametrize(
        "slope_exp, weights",
        [
            (0.0, [1 / 8, 1 / 8, 1 / 8, 1 / 8, 1 / 8, 1 / 8, 1 / 8, 1 / 8]),
            (2.0, [2 / 12, 1 / 12, 2 / 12, 1 / 12, 1 / 12, 2 / 12, 1 / 12, 2 / 12]),
        ],
    )
    def test_flow_partitions(self, slope_exp, weights) -> None:
        # test on a 3x3 tiny grid with base levels at all border nodes
        grid = RasterGrid([3, 3], [1.0, 1.0], NodeStatus.FIXED_VALUE)

        sqrt2 = np.sqrt(2)
        elevation = np.array(
            [[0.0, sqrt2, 0.0], [sqrt2, sqrt2 * 2, sqrt2], [0.0, sqrt2, 0.0]]
        )
        router = MultiFlowRouter(slope_exp)
        flow_graph = FlowGraph(grid, [router])
        flow_graph.update_routes(elevation)

        npt.assert_allclose(
            flow_graph.impl().receivers_weight[4],
            weights,
        )


@pytest.mark.parametrize(
    "router", [SingleFlowRouter(), MultiFlowRouter(0.0), MultiFlowRouter(2.0)]
)
def test_conservation_area(router) -> None:
    # High level test: conservative flow routing
    nrows = 10
    ncols = 8

    # bottom border base level
    bottom_base_level = [
        NodeStatus.CORE,
        NodeStatus.CORE,
        NodeStatus.CORE,
        NodeStatus.FIXED_VALUE,
    ]
    grid = RasterGrid([nrows, ncols], [1.0, 1.0], bottom_base_level)

    # avoid closed depressions (all flow must reach bottom border nodes)
    flow_graph = FlowGraph(grid, [PFloodSinkResolver(), router])

    # planar surface tilted along the y-axis + random perturbations
    elevation = np.random.uniform(size=grid.shape) + np.arange(nrows)[:, None] * 2

    flow_graph.update_routes(elevation)
    drainage_area = flow_graph.accumulate(1.0)

    # assumes grid cell uniform area is 1
    assert abs(np.sum(drainage_area[-1]) - grid.size) < 1e-5


@pytest.mark.parametrize(
    "router", [SingleFlowRouter(), MultiFlowRouter(0.0), MultiFlowRouter(2.0)]
)
def test_monotonic_dfs(router) -> None:
    # High level test: monotonic elevation for dfs indices
    nrows = 10
    ncols = 8

    grid = RasterGrid([nrows, ncols], [1.0, 1.0], NodeStatus.FIXED_VALUE)

    # this test requires no closed depressions (fill them with tiny slope)
    flow_graph = FlowGraph(grid, [PFloodSinkResolver(), router])

    elevation = np.random.uniform(size=grid.shape)
    filled_elevation = flow_graph.update_routes(elevation)
    filled_elevation_flat = filled_elevation.ravel()

    receivers_count = flow_graph.impl().receivers_count
    receivers = flow_graph.impl().receivers
    dfs_indices = flow_graph.impl().dfs_indices

    # traverse the graph, check that elevation increases if
    # previous node is a direct receiver of current visited node
    # (dfs_indices always in bottom->up direction)
    for i, inode in enumerate(dfs_indices[1:]):
        iprev = dfs_indices[i - 1]
        if iprev in receivers[inode][: receivers_count[inode]]:
            assert filled_elevation_flat[inode] >= filled_elevation_flat[iprev]



import numpy as np
import pytest

from fastscapelib.flow import (
    FlowDirection,
    FlowGraph,
    FlowOperator,
    MSTMethod,
    MSTRouteMethod,
    MSTSinkResolver,
    PFloodSinkResolver,
    SingleFlowRouter,
)
from fastscapelib.grid import NodeStatus, RasterGrid


@pytest.fixture
def grid():
    bs = [
        NodeStatus.CORE,
        NodeStatus.CORE,
        NodeStatus.CORE,
        NodeStatus.FIXED_VALUE,
    ]
    grid = RasterGrid([5, 5], [1, 1], bs)

    yield grid


@pytest.fixture
def elevation():
    # planar surface tilted along the y-axis
    # + one closed depression with pit at row/col index (1, 3)
    #   and pass between (2, 2) and (3, 1)
    elevation = np.array(
        [
            [1.00, 1.00, 1.00, 1.00, 1.00],
            [0.70, 0.70, 0.70, 0.10, 0.70],
            [0.50, 0.50, 0.15, 0.50, 0.50],
            [0.20, 0.19, 0.20, 0.20, 0.20],
            [0.00, 0.00, 0.00, 0.00, 0.00],
        ]
    )

    yield elevation


def test_no_sink_resolver(grid, elevation) -> None:
    graph = FlowGraph(grid, [SingleFlowRouter()])

    new_elevation = graph.update_routes(elevation)

    # unchanged elevation
    np.testing.assert_array_equal(new_elevation, elevation)

    # flow trapped in pit
    pit_idx_flat = 8
    assert graph.impl().receivers[pit_idx_flat, 0] == pit_idx_flat


def test_pflood_sink_resolver_attrs() -> None:
    assert PFloodSinkResolver.graph_updated is False
    assert PFloodSinkResolver.elevation_updated is True
    assert PFloodSinkResolver.in_flowdir == FlowDirection.UNDEFINED
    assert PFloodSinkResolver.out_flowdir == FlowDirection.UNDEFINED


def test_pflood_sink_resolver(grid, elevation) -> None:
    resolver = PFloodSinkResolver()
    assert isinstance(resolver, FlowOperator)
    assert resolver.name == "pflood_sink_resolver"
    assert repr(resolver) == "PFloodSinkResolver"

    graph = FlowGraph(grid, [resolver, SingleFlowRouter()])

    new_elevation = graph.update_routes(elevation)

    # filled elevation (tiny slope)
    mask_filled = np.zeros_like(elevation, dtype=bool)
    mask_filled[[1, 2], [3, 2]] = True
    np.testing.assert_array_equal(new_elevation[~mask_filled], elevation[~mask_filled])
    assert new_elevation[2, 2] > new_elevation[3, 1]
    assert new_elevation[1, 3] > new_elevation[2, 2]

    # resolved flow
    assert graph.impl().receivers[8, 0] == 12
    assert graph.impl().receivers_distance[8, 0] == np.sqrt(2)
    assert graph.impl().receivers[12, 0] == 16
    assert graph.impl().receivers_distance[12, 0] == np.sqrt(2)


class TestMSTSinkResolver:
    def test_class_attrs(self) -> None:
        assert MSTSinkResolver.graph_updated is True
        assert MSTSinkResolver.elevation_updated is True
        assert MSTSinkResolver.in_flowdir == FlowDirection.SINGLE
        assert MSTSinkResolver.out_flowdir == FlowDirection.SINGLE

    def test_constructor(self) -> None:
        resolver = MSTSinkResolver()
        assert isinstance(resolver, FlowOperator)
        assert resolver.basin_method == MSTMethod.KRUSKAL
        assert resolver.route_method == MSTRouteMethod.CARVE
        assert resolver.name == "mst_sink_resolver"
        assert repr(resolver) == "MSTSinkResolver (basin=kruskal, route=carve)"

        # read-write attributes
        resolver.basin_method = MSTMethod.BORUVKA
        assert resolver.basin_method == MSTMethod.BORUVKA

        resolver.route_method = MSTRouteMethod.BASIC
        assert resolver.route_method == MSTRouteMethod.BASIC

        # constructor with arguments
        resolver2 = MSTSinkResolver(MSTMethod.BORUVKA, MSTRouteMethod.BASIC)
        assert resolver2.basin_method == MSTMethod.BORUVKA
        assert resolver2.route_method == MSTRouteMethod.BASIC

    @pytest.mark.parametrize("mst_method", [MSTMethod.KRUSKAL, MSTMethod.BORUVKA])
    def test_resolve_basic(self, grid, elevation, mst_method) -> None:
        resolver = MSTSinkResolver(mst_method, MSTRouteMethod.BASIC)
        graph = FlowGraph(grid, [SingleFlowRouter(), resolver])

        new_elevation = graph.update_routes(elevation)

        # filled elevation (tiny slope)
        mask_filled = np.zeros_like(elevation, dtype=bool)
        mask_filled[[1, 2], [3, 2]] = True
        np.testing.assert_array_equal(
            new_elevation[~mask_filled], elevation[~mask_filled]
        )
        # pit node (1, 3) is elevated first above the pass node (3, 1)
        # then sink node (2, 2) is elevated above the pit node
        assert new_elevation[1, 3] > new_elevation[3, 1]
        assert new_elevation[2, 2] > new_elevation[1, 3]

        # only pit node is updated
        assert graph.impl().receivers[8, 0] == 16
        assert graph.impl().receivers_distance[8, 0] > 1e99  # infinity
        assert graph.impl().receivers[12, 0] == 8
        assert graph.impl().receivers_distance[12, 0] == np.sqrt(2)

    @pytest.mark.parametrize("mst_method", [MSTMethod.KRUSKAL, MSTMethod.BORUVKA])
    def test_resolve_carve(self, grid, elevation, mst_method) -> None:
        resolver = MSTSinkResolver(mst_method, MSTRouteMethod.CARVE)
        graph = FlowGraph(grid, [SingleFlowRouter(), resolver])

        new_elevation = graph.update_routes(elevation)

        # filled elevation (tiny slope)
        mask_filled = np.zeros_like(elevation, dtype=bool)
        mask_filled[[1, 2], [3, 2]] = True
        np.testing.assert_array_equal(
            new_elevation[~mask_filled], elevation[~mask_filled]
        )
        assert new_elevation[2, 2] > new_elevation[3, 1]
        assert new_elevation[1, 3] > new_elevation[2, 2]

        # resolved flow
        assert graph.impl().receivers[8, 0] == 12
        assert graph.impl().receivers_distance[8] == np.sqrt(2)
        assert graph.impl().receivers[12, 0] == 16
        assert graph.impl().receivers_distance[12, 0] == np.sqrt(2)


@pytest.mark.parametrize(
    "resolver",
    [
        PFloodSinkResolver(),
        MSTSinkResolver(MSTMethod.KRUSKAL, MSTRouteMethod.BASIC),
        MSTSinkResolver(MSTMethod.KRUSKAL, MSTRouteMethod.CARVE),
        MSTSinkResolver(MSTMethod.BORUVKA, MSTRouteMethod.BASIC),
        MSTSinkResolver(MSTMethod.BORUVKA, MSTRouteMethod.CARVE),
    ],
)
def test_conservation_drainage_area(resolver) -> None:
    # High-level test for sink resolvers that checks the conservation of
    # drainage area (sum of drainage area at domain fixed bounaries = total
    # domain area)

    shape = (101, 101)
    grid = RasterGrid([101, 101], [1, 1], NodeStatus.FIXED_VALUE)

    elevation = np.random.uniform(size=shape)

    if isinstance(resolver, PFloodSinkResolver):
        graph = FlowGraph(grid, [resolver, SingleFlowRouter()])
    else:
        graph = FlowGraph(grid, [SingleFlowRouter(), resolver])

    graph.update_routes(elevation)
    drainage_area = graph.accumulate(1.0)

    actual = drainage_area[1:-1, [0, -1]].sum() + drainage_area[[0, -1]].sum()
    expected = elevation.size

    assert actual == expected


@pytest.mark.parametrize(
    "resolver",
    [
        PFloodSinkResolver(),
        MSTSinkResolver(MSTMethod.KRUSKAL, MSTRouteMethod.BASIC),
        MSTSinkResolver(MSTMethod.KRUSKAL, MSTRouteMethod.CARVE),
        MSTSinkResolver(MSTMethod.BORUVKA, MSTRouteMethod.BASIC),
        MSTSinkResolver(MSTMethod.BORUVKA, MSTRouteMethod.CARVE),
    ],
)
def test_nb_of_basins(resolver) -> None:
    # High-level test for sink resolvers that checks that the total number of
    # basins equals the number of (fixed value) boundary nodes at grid borders
    # (i.e., only outer basins)

    shape = (101, 101)
    grid = RasterGrid([101, 101], [1, 1], NodeStatus.FIXED_VALUE)

    elevation = np.random.uniform(size=shape)

    if isinstance(resolver, PFloodSinkResolver):
        graph = FlowGraph(grid, [resolver, SingleFlowRouter()])
    else:
        graph = FlowGraph(grid, [SingleFlowRouter(), resolver])

    graph.update_routes(elevation)
    basins = graph.basins()
    nb_basins = basins.max() + 1

    nb_border_nodes = (shape[0] + shape[1]) * 2 - 4

    assert nb_basins == nb_border_nodes



import numpy as np
import numpy.testing as npt
import pytest

from fastscapelib.grid import (
    Neighbor,
    NodeStatus,
    RasterBoundaryStatus,
    RasterGrid,
    RasterNeighbor,
)


class TestRasterBoundaryStatus:
    def setup_method(self) -> None:
        self.bs1 = RasterBoundaryStatus(
            [
                NodeStatus.CORE,
                NodeStatus.FIXED_VALUE,
                NodeStatus.CORE,
                NodeStatus.FIXED_VALUE,
            ]
        )
        self.bs2 = RasterBoundaryStatus(NodeStatus.CORE)
        self.bs3 = RasterBoundaryStatus(NodeStatus.LOOPED)

    def test__init__(self) -> None:
        # call the setup_method
        pass

    def test_left(self) -> None:
        assert self.bs1.left == NodeStatus.CORE
        assert self.bs2.left == NodeStatus.CORE
        assert self.bs3.left == NodeStatus.LOOPED

    def test_right(self) -> None:
        assert self.bs1.right == NodeStatus.FIXED_VALUE
        assert self.bs2.right == NodeStatus.CORE
        assert self.bs3.right == NodeStatus.LOOPED

    def test_bottom(self) -> None:
        assert self.bs1.bottom == NodeStatus.FIXED_VALUE
        assert self.bs2.bottom == NodeStatus.CORE
        assert self.bs3.bottom == NodeStatus.LOOPED

    def test_top(self) -> None:
        assert self.bs1.top == NodeStatus.CORE
        assert self.bs2.top == NodeStatus.CORE
        assert self.bs3.top == NodeStatus.LOOPED

    def test_is_horizontal_looped(self) -> None:
        assert not self.bs1.is_horizontal_looped
        assert not self.bs2.is_horizontal_looped
        assert self.bs3.is_horizontal_looped

    def test_is_vertical_looped(self) -> None:
        assert not self.bs1.is_vertical_looped
        assert not self.bs2.is_vertical_looped
        assert self.bs3.is_vertical_looped


class TestRasterNeighbor:
    def setup_method(self) -> None:
        self.n = RasterNeighbor(5, 0, 5, 1.35, NodeStatus.CORE)

    def test__init__(self) -> None:
        # call the setup_method
        assert self.n == RasterNeighbor(5, 0, 5, 1.35, NodeStatus.CORE)
        pass

    def test_flatten_idx(self) -> None:
        assert self.n.flatten_idx == 5

        self.n.flatten_idx = 8
        assert self.n.flatten_idx == 8

    def test_row(self) -> None:
        assert self.n.row == 0

        self.n.row = 3
        assert self.n.row == 3

    def test_col(self) -> None:
        assert self.n.col == 5

        self.n.col = 8
        assert self.n.col == 8

    def test_distance(self) -> None:
        assert self.n.distance == 1.35

        self.n.distance = 1.38
        assert self.n.distance == 1.38

    def test_status(self) -> None:
        assert self.n.status == NodeStatus.CORE

        self.n.status = NodeStatus.FIXED_VALUE
        assert self.n.status == NodeStatus.FIXED_VALUE


class TestRasterGrid:
    def setup_method(self) -> None:
        self.bs = bs = RasterBoundaryStatus(NodeStatus.FIXED_VALUE)
        self.g = RasterGrid([5, 10], [2.2, 2.4], bs, {(0, 5): NodeStatus.FIXED_VALUE})

    def test_static_props(self) -> None:
        assert RasterGrid.is_structured is True
        assert RasterGrid.is_uniform is True
        assert RasterGrid.n_neighbors_max == 8

    def test_constructor(self) -> None:
        g1 = RasterGrid([10, 10], [2.3, 2.1], self.bs, {(0, 5): NodeStatus.FIXED_VALUE})
        assert g1.size == 100
        npt.assert_almost_equal(g1.spacing, [2.3, 2.1])
        npt.assert_almost_equal(g1.length, [20.7, 18.9])

        with pytest.raises(IndexError):
            RasterGrid(
                [5, 10],
                [2.2, 2.4],
                self.bs,
                {(20, 255): NodeStatus.FIXED_VALUE},
            )

    @pytest.mark.parametrize(
        "bounds_status",
        [
            NodeStatus.FIXED_VALUE,
            [NodeStatus.FIXED_VALUE] * 4,
            RasterBoundaryStatus(NodeStatus.FIXED_VALUE),
        ],
    )
    def test_constructor_bounds_status(self, bounds_status) -> None:
        g = RasterGrid([2, 2], [2.0, 2.0], bounds_status)
        expected = np.ones((2, 2)) * NodeStatus.FIXED_VALUE.value
        npt.assert_array_equal(g.nodes_status(), expected)

    def test_from_length(self) -> None:
        g = RasterGrid.from_length(
            [11, 11], np.r_[23, 21], self.bs, {(0, 5): NodeStatus.FIXED_VALUE}
        )
        assert g.size == 121
        npt.assert_almost_equal(g.spacing, [2.3, 2.1])
        npt.assert_almost_equal(g.length, [23.0, 21.0])

    @pytest.mark.parametrize(
        "bounds_status",
        [
            NodeStatus.FIXED_VALUE,
            [NodeStatus.FIXED_VALUE] * 4,
            RasterBoundaryStatus(NodeStatus.FIXED_VALUE),
        ],
    )
    def test_from_length_bounds_status(self, bounds_status) -> None:
        g = RasterGrid.from_length([2, 2], [20.0, 20.0], bounds_status)
        expected = np.ones((2, 2)) * NodeStatus.FIXED_VALUE.value
        npt.assert_array_equal(g.nodes_status(), expected)

    def test_size(self) -> None:
        assert self.g.size == 50

    def test_shape(self) -> None:
        npt.assert_equal(self.g.shape, [5, 10])

    def test_spacing(self) -> None:
        npt.assert_equal(self.g.spacing, [2.2, 2.4])

    def test_nodes_indices(self) -> None:
        grid = RasterGrid([3, 3], [2.0, 2.0], self.bs)
        npt.assert_equal(grid.nodes_indices(), np.arange(grid.size))
        npt.assert_equal(
            grid.nodes_indices(NodeStatus.FIXED_VALUE), [0, 1, 2, 3, 5, 6, 7, 8]
        )
        npt.assert_equal(grid.nodes_indices(NodeStatus.CORE), [4])
        assert not len(grid.nodes_indices(NodeStatus.FIXED_GRADIENT))

    def test_nodes_status(self) -> None:
        grid = RasterGrid([3, 3], [2.0, 2.0], self.bs)

        npt.assert_equal(grid.nodes_status(), [[1, 1, 1], [1, 0, 1], [1, 1, 1]])
        # should return a copy
        assert not np.shares_memory(grid.nodes_status(), grid.nodes_status())

        assert grid.nodes_status(0) == NodeStatus.FIXED_VALUE
        assert grid.nodes_status(4) == NodeStatus.CORE

    def test_nodes_areas(self) -> None:
        area = 2.2 * 2.4
        assert self.g.nodes_areas(0) == area
        assert self.g.nodes_areas(10) == area
        npt.assert_equal(self.g.nodes_areas(), np.full(self.g.shape, area))

    def test_neighbors_count(self) -> None:
        assert self.g.neighbors_count(0) == 3
        assert self.g.neighbors_count(15) == 8

    def test_neighbors_indices(self) -> None:
        npt.assert_equal(self.g.neighbors_indices(0), np.array([1, 10, 11]))
        npt.assert_equal(
            self.g.neighbors_indices(15), np.array([4, 5, 6, 14, 16, 24, 25, 26])
        )

        with pytest.raises(IndexError, match="grid index out of range"):
            self.g.neighbors(51)

    def test_neighbors_raster_indices(self) -> None:
        assert self.g.neighbors_indices(0, 0) == [(0, 1), (1, 0), (1, 1)]
        assert self.g.neighbors_indices(1, 5) == [
            (0, 4),
            (0, 5),
            (0, 6),
            (1, 4),
            (1, 6),
            (2, 4),
            (2, 5),
            (2, 6),
        ]

        with pytest.raises(IndexError, match="grid index out of range"):
            self.g.neighbors(10, 10)

    def test_neighbors_distances(self) -> None:
        dist_diag = np.sqrt(2.2**2 + 2.4**2)
        npt.assert_equal(self.g.neighbors_distances(0), np.array([2.4, 2.2, dist_diag]))
        npt.assert_equal(
            self.g.neighbors_distances(15),
            np.array([dist_diag, 2.2, dist_diag, 2.4, 2.4, dist_diag, 2.2, dist_diag]),
        )

        with pytest.raises(IndexError, match="grid index out of range"):
            self.g.neighbors(51)

    def test_neighbors(self) -> None:
        dist_diag = np.sqrt(2.2**2 + 2.4**2)
        assert self.g.neighbors(0) == [
            Neighbor(1, 2.4, NodeStatus.FIXED_VALUE),
            Neighbor(10, 2.2, NodeStatus.FIXED_VALUE),
            Neighbor(11, dist_diag, NodeStatus.CORE),
        ]

        with pytest.raises(IndexError, match="grid index out of range"):
            self.g.neighbors(51)

    def test_raster_neighbors(self) -> None:
        dist_diag = np.sqrt(2.2**2 + 2.4**2)

        assert self.g.neighbors(0, 0) == [
            RasterNeighbor(1, 0, 1, 2.4, NodeStatus.FIXED_VALUE),
            RasterNeighbor(10, 1, 0, 2.2, NodeStatus.FIXED_VALUE),
            RasterNeighbor(11, 1, 1, dist_diag, NodeStatus.CORE),
        ]

        with pytest.raises(IndexError, match="grid index out of range"):
            self.g.neighbors(10, 10)



from textwrap import dedent

import numpy as np
import numpy.testing as npt
import pytest

from fastscapelib.flow import (
    FlowGraph,
    FlowSnapshot,
    MultiFlowRouter,
    PFloodSinkResolver,
    SingleFlowRouter,
)
from fastscapelib.grid import NodeStatus, ProfileGrid, RasterGrid


class TestFlowGraph:
    def test___init__(self) -> None:
        profile_grid = ProfileGrid(8, 2.2, NodeStatus.FIXED_VALUE)
        raster_grid = RasterGrid([5, 10], [2.2, 2.4], NodeStatus.FIXED_VALUE)

        FlowGraph(profile_grid, [SingleFlowRouter()])
        FlowGraph(raster_grid, [PFloodSinkResolver(), SingleFlowRouter()])

        with pytest.raises(
            ValueError, match="must have at least one operator that updates"
        ):
            FlowGraph(raster_grid, [])

        with pytest.raises(TypeError, match="invalid flow operator"):
            FlowGraph(raster_grid, ["not a flow operator"])  # type: ignore[list-item]

    def test_base_levels(self) -> None:
        raster_grid = RasterGrid([3, 3], [1.0, 1.0], NodeStatus.FIXED_VALUE)
        flow_graph = FlowGraph(raster_grid, [SingleFlowRouter()])

        actual = flow_graph.base_levels
        assert set(actual) == {0, 1, 2, 3, 5, 6, 7, 8}

        flow_graph.base_levels = [3]
        assert flow_graph.base_levels == [3]

    def test_mask(self) -> None:
        raster_grid = RasterGrid([3, 3], [1.0, 1.0], NodeStatus.FIXED_VALUE)
        flow_graph = FlowGraph(raster_grid, [SingleFlowRouter()])

        # uninitialized mask
        npt.assert_array_equal(flow_graph.mask, np.array([], dtype=np.bool_))

        flow_graph.mask = np.ones(raster_grid.shape, dtype=np.bool_)
        npt.assert_array_equal(
            flow_graph.mask, np.ones(raster_grid.shape, dtype=np.bool_)
        )

    def test_operators(self) -> None:
        grid = ProfileGrid(8, 2.2, NodeStatus.FIXED_VALUE)

        resolver = PFloodSinkResolver()
        router = SingleFlowRouter()

        flow_graph = FlowGraph(grid, [resolver, router])

        assert flow_graph.operators == [resolver, router]

    def test_single_flow(self) -> None:
        grid = ProfileGrid(8, 2.2, NodeStatus.FIXED_VALUE)

        flow_graph = FlowGraph(grid, [SingleFlowRouter()])
        assert flow_graph.single_flow is True

        flow_graph = FlowGraph(grid, [MultiFlowRouter(1.0)])
        assert flow_graph.single_flow is False

    def test_repr(self) -> None:
        grid = ProfileGrid(8, 2.2, NodeStatus.FIXED_VALUE)
        flow_graph = FlowGraph(grid, [SingleFlowRouter(), FlowSnapshot("test")])

        actual = repr(flow_graph)
        expected = """
        <FlowGraph (8 nodes)>
        Operators:
            SingleFlowRouter
            FlowSnapshot 'test' (graph=True, elevation=False)
        """
        assert actual == dedent(expected[1:])

        actual = repr(flow_graph.graph_snapshot("test"))
        expected = """
        <FlowGraph (8 nodes)>
        Operators:
            *empty*
        """
        assert actual == dedent(expected[1:])

    def test_update_routes(self) -> None:
        grid = ProfileGrid(8, 2.2, NodeStatus.FIXED_VALUE)
        flow_graph = FlowGraph(grid, [SingleFlowRouter()])

        # pit at 3rd node
        elevation = np.array([0.0, 0.2, 0.1, 0.2, 0.4, 0.6, 0.3, 0.0])

        new_elevation = flow_graph.update_routes(elevation)

        npt.assert_array_equal(elevation, new_elevation)

        npt.assert_equal(
            flow_graph.impl().receivers[:, 0], np.array([0, 0, 2, 2, 3, 6, 7, 7])
        )
        npt.assert_equal(flow_graph.impl().receivers_count, np.ones(elevation.size))
        npt.assert_equal(
            flow_graph.impl().receivers_weight[:, 0], np.ones(elevation.size)
        )

        actual = flow_graph.impl().donors
        m = np.iinfo(actual.dtype).max
        npt.assert_equal(
            actual,
            np.array(
                [
                    [1, m, m],
                    [m, m, m],
                    [2, 3, m],
                    [4, m, m],
                    [m, m, m],
                    [m, m, m],
                    [5, m, m],
                    [6, m, m],
                ]
            ),
        )
        npt.assert_equal(
            flow_graph.impl().donors_count, np.array([1, 0, 2, 1, 0, 0, 1, 1])
        )

    def test_accumulate_basins(self) -> None:
        # --- test profile grid
        grid = ProfileGrid(8, 2.0, NodeStatus.FIXED_VALUE)
        flow_graph = FlowGraph(grid, [SingleFlowRouter()])

        # pit at 3rd node
        elevation = np.array([0.0, 0.2, 0.1, 0.2, 0.4, 0.6, 0.3, 0.0])

        new_elevation = flow_graph.update_routes(elevation)

        npt.assert_array_equal(elevation, new_elevation)

        acc = np.empty_like(elevation)
        src = np.ones_like(elevation)
        expected = np.array([4.0, 2.0, 6.0, 4.0, 2.0, 2.0, 4.0, 6.0])

        flow_graph.accumulate(acc, 1.0)
        npt.assert_array_equal(acc, expected)

        flow_graph.accumulate(acc, src)
        npt.assert_array_equal(acc, expected)

        npt.assert_almost_equal(flow_graph.accumulate(1.0), expected)
        npt.assert_almost_equal(flow_graph.accumulate(src), expected)

        # bottom border base-level
        bottom_base_level = [
            NodeStatus.CORE,
            NodeStatus.CORE,
            NodeStatus.CORE,
            NodeStatus.FIXED_VALUE,
        ]

        # --- test raster grid
        rgrid = RasterGrid([4, 4], [1.0, 1.0], bottom_base_level)

        flow_graph = FlowGraph(rgrid, [SingleFlowRouter()])

        # planar surface tilted along the y-axis + small carved channel
        elevation = np.array(
            [
                [0.6, 0.6, 0.6, 0.6],
                [0.4, 0.4, 0.4, 0.4],
                [0.2, 0.2, 0.2, 0.2],
                [0.1, 0.0, 0.1, 0.1],
            ]
        )

        new_elevation = flow_graph.update_routes(elevation)
        npt.assert_array_equal(elevation, new_elevation)

        acc = np.empty_like(elevation)
        src = np.ones_like(elevation)
        expected = np.array(
            [
                [1.0, 1.0, 1.0, 1.0],
                [2.0, 2.0, 2.0, 2.0],
                [3.0, 3.0, 3.0, 3.0],
                [1.0, 10.0, 1.0, 4.0],
            ]
        )

        flow_graph.accumulate(acc, 1.0)
        npt.assert_array_equal(acc, expected)

        flow_graph.accumulate(acc, src)
        npt.assert_array_equal(acc, expected)

        npt.assert_almost_equal(flow_graph.accumulate(1.0), expected)
        npt.assert_almost_equal(flow_graph.accumulate(src), expected)

        expected = np.array(
            [
                [1, 1, 1, 3],
                [1, 1, 1, 3],
                [1, 1, 1, 3],
                [0, 1, 2, 3],
            ]
        )
        npt.assert_array_equal(flow_graph.basins(), expected)
        npt.assert_array_equal(flow_graph.basins().flatten(), flow_graph.impl().basins)

        # test basins mask
        mask = np.array(
            [
                [False, False, False, True],
                [False, False, False, True],
                [False, False, False, True],
                [False, False, False, True],
            ]
        )

        flow_graph.mask = mask
        flow_graph.update_routes(elevation)
        actual = flow_graph.basins()
        no_basin = np.iinfo(actual.dtype).max
        expected = np.array(
            [
                [1, 1, 1, no_basin],
                [1, 1, 1, no_basin],
                [1, 1, 1, no_basin],
                [0, 1, 2, no_basin],
            ]
        )
        npt.assert_array_equal(actual, expected)



import numpy as np
import numpy.testing as npt
import pytest

from fastscapelib.grid import (
    Neighbor,
    Node,
    NodeStatus,
    ProfileBoundaryStatus,
    ProfileGrid,
)


class TestProfileBoundaryStatus:
    def setup_method(self) -> None:
        self.bs1 = ProfileBoundaryStatus([NodeStatus.CORE, NodeStatus.FIXED_VALUE])
        self.bs2 = ProfileBoundaryStatus(NodeStatus.FIXED_VALUE, NodeStatus.CORE)
        self.bs3 = ProfileBoundaryStatus(NodeStatus.CORE)
        self.bs4 = ProfileBoundaryStatus(NodeStatus.LOOPED)

    def test__init__(self) -> None:
        # call the setup_method
        pass

    def test_left(self) -> None:
        assert self.bs1.left == NodeStatus.CORE
        assert self.bs2.left == NodeStatus.FIXED_VALUE
        assert self.bs3.left == NodeStatus.CORE
        assert self.bs4.left == NodeStatus.LOOPED

    def test_right(self) -> None:
        assert self.bs1.right == NodeStatus.FIXED_VALUE
        assert self.bs2.right == NodeStatus.CORE
        assert self.bs3.right == NodeStatus.CORE
        assert self.bs4.right == NodeStatus.LOOPED

    def test_is_horizontal_looped(self) -> None:
        assert not self.bs1.is_horizontal_looped
        assert not self.bs2.is_horizontal_looped
        assert not self.bs3.is_horizontal_looped
        assert self.bs4.is_horizontal_looped


class TestNeighbor:
    def setup_method(self) -> None:
        self.n = Neighbor(5, 1.35, NodeStatus.CORE)

    def test__init__(self) -> None:
        # call the setup_method
        assert self.n == Neighbor(5, 1.35, NodeStatus.CORE)
        pass

    def test_idx(self) -> None:
        assert self.n.idx == 5

        self.n.idx = 8
        assert self.n.idx == 8

    def test_distance(self) -> None:
        assert self.n.distance == 1.35

        self.n.distance = 1.38
        assert self.n.distance == 1.38

    def test_status(self) -> None:
        assert self.n.status == NodeStatus.CORE

        self.n.status = NodeStatus.FIXED_VALUE
        assert self.n.status == NodeStatus.FIXED_VALUE


class TestProfileGrid:
    def setup_method(self) -> None:
        self.bs = [NodeStatus.FIXED_VALUE] * 2
        self.g = ProfileGrid(10, 2.2, self.bs, nodes_status={5: NodeStatus.FIXED_VALUE})

    def test_static_props(self) -> None:
        assert ProfileGrid.is_structured is True
        assert ProfileGrid.is_uniform is True
        assert ProfileGrid.n_neighbors_max == 2

    def test_constructor(self) -> None:
        g = ProfileGrid(10, 2, self.bs, nodes_status={5: NodeStatus.FIXED_VALUE})
        assert g.size == 10
        assert g.spacing == 2.0
        assert g.length == 18.0

        g = ProfileGrid(
            15,
            3.0,
            ProfileBoundaryStatus(self.bs),
            {5: NodeStatus.FIXED_VALUE},
        )
        assert g.size == 15
        assert g.spacing == 3.0
        assert g.length == 42.0

        with pytest.raises(IndexError):
            ProfileGrid(10, 2, self.bs, {15: NodeStatus.FIXED_VALUE})

    @pytest.mark.parametrize(
        "bounds_status",
        [
            NodeStatus.FIXED_VALUE,
            [NodeStatus.FIXED_VALUE] * 2,
            ProfileBoundaryStatus(NodeStatus.FIXED_VALUE),
        ],
    )
    def test_constructor_bounds_status(self, bounds_status) -> None:
        g = ProfileGrid(10, 2, bounds_status)
        assert g.nodes_status(0) == g.nodes_status(9) == NodeStatus.FIXED_VALUE

    def test_from_length(self) -> None:
        g = ProfileGrid.from_length(
            11,
            20.0,
            ProfileBoundaryStatus(self.bs),
            {5: NodeStatus.FIXED_VALUE},
        )
        assert g.size == 11
        assert g.shape == [11]
        assert g.spacing == 2.0
        assert g.length == 20.0

    @pytest.mark.parametrize(
        "bounds_status",
        [
            NodeStatus.FIXED_VALUE,
            [NodeStatus.FIXED_VALUE] * 2,
            ProfileBoundaryStatus(NodeStatus.FIXED_VALUE),
        ],
    )
    def test_from_length_bounds_status(self, bounds_status) -> None:
        g = ProfileGrid.from_length(10, 20.0, bounds_status)
        assert g.nodes_status(0) == g.nodes_status(9) == NodeStatus.FIXED_VALUE

    def test_size(self) -> None:
        assert self.g.size == 10

    def test_spacing(self) -> None:
        assert self.g.spacing == 2.2

    def test_shape(self) -> None:
        npt.assert_almost_equal(self.g.shape, np.r_[10])

    def test_length(self) -> None:
        assert self.g.length == 19.8

    def test_nodes_indices(self) -> None:
        npt.assert_equal(self.g.nodes_indices(), np.arange(self.g.size))
        npt.assert_equal(self.g.nodes_indices(NodeStatus.FIXED_VALUE), [0, 5, 9])
        npt.assert_equal(self.g.nodes_indices(NodeStatus.CORE), [1, 2, 3, 4, 6, 7, 8])
        assert not len(self.g.nodes_indices(NodeStatus.FIXED_GRADIENT))

    def test_nodes_status(self) -> None:
        npt.assert_equal(
            self.g.nodes_status(), np.array([1, 0, 0, 0, 0, 1, 0, 0, 0, 1])
        )
        # should return a copy
        assert not np.shares_memory(self.g.nodes_status(), self.g.nodes_status())

        assert self.g.nodes_status(0) == NodeStatus.FIXED_VALUE
        assert self.g.nodes_status(1) == NodeStatus.CORE

    def test_nodes_areas(self) -> None:
        assert self.g.nodes_areas(0) == 2.2
        assert self.g.nodes_areas(2) == 2.2
        npt.assert_equal(self.g.nodes_areas(), np.full(self.g.size, 2.2))

    def test_neighbors_count(self) -> None:
        assert self.g.neighbors_count(0) == 1
        assert self.g.neighbors_count(5) == 2

    def test_neighbors_indices(self) -> None:
        npt.assert_equal(self.g.neighbors_indices(0), np.array([1]))
        npt.assert_equal(self.g.neighbors_indices(5), np.array([4, 6]))

        with pytest.raises(IndexError, match="grid index out of range"):
            self.g.neighbors(11)

    def test_neighbors_distances(self) -> None:
        npt.assert_equal(self.g.neighbors_distances(0), np.array([2.2]))
        npt.assert_equal(self.g.neighbors_distances(5), np.array([2.2, 2.2]))

        with pytest.raises(IndexError, match="grid index out of range"):
            self.g.neighbors(11)

    def test_neighbors(self) -> None:
        assert self.g.neighbors(0) == [Neighbor(1, 2.2, NodeStatus.CORE)]
        assert self.g.neighbors(1) == [
            Neighbor(0, 2.2, NodeStatus.FIXED_VALUE),
            Neighbor(2, 2.2, NodeStatus.CORE),
        ]
        assert self.g.neighbors(6) == [
            Neighbor(5, 2.2, NodeStatus.FIXED_VALUE),
            Neighbor(7, 2.2, NodeStatus.CORE),
        ]

        with pytest.raises(IndexError, match="grid index out of range"):
            self.g.neighbors(11)



from _fastscapelib_py.flow import (  # type: ignore[import]
    FlowDirection,
    FlowGraph,
    FlowGraphImpl,
    FlowOperator,
    FlowSnapshot,
    MSTMethod,
    MSTRouteMethod,
    MSTSinkResolver,
    MultiFlowRouter,
    PFloodSinkResolver,
    SingleFlowRouter,
)

__all__ = [
    "FlowDirection",
    "FlowGraph",
    "FlowGraphImpl",
    "FlowOperator",
    "FlowOperator",
    "FlowSnapshot",
    "MSTMethod",
    "MSTRouteMethod",
    "MSTSinkResolver",
    "MultiFlowRouter",
    "PFloodSinkResolver",
    "SingleFlowRouter",
]



from _fastscapelib_py import __version__  # type: ignore[import]

from fastscapelib.eroders import DiffusionADIEroder, SPLEroder  # type: ignore[import]
from fastscapelib.flow import (  # type: ignore[import]
    FlowDirection,
    FlowGraph,
    FlowGraphImpl,
    FlowOperator,
    FlowSnapshot,
    MSTMethod,
    MSTRouteMethod,
    MSTSinkResolver,
    MultiFlowRouter,
    PFloodSinkResolver,
    SingleFlowRouter,
)
from fastscapelib.grid import (  # type: ignore[import]
    Neighbor,
    Node,
    NodeStatus,
    ProfileBoundaryStatus,
    ProfileGrid,
    RasterBoundaryStatus,
    RasterGrid,
    RasterNeighbor,
    RasterNode,
    TriMesh,
)

__all__ = [
    "__version__",
    "DiffusionADIEroder",
    "SPLEroder",
    "Neighbor",
    "Node",
    "NodeStatus",
    "ProfileBoundaryStatus",
    "ProfileGrid",
    "RasterBoundaryStatus",
    "RasterGrid",
    "RasterNeighbor",
    "RasterNode",
    "TriMesh",
    "FlowDirection",
    "FlowGraph",
    "FlowGraphImpl",
    "FlowOperator",
    "FlowOperator",
    "FlowSnapshot",
    "MSTMethod",
    "MSTRouteMethod",
    "MSTSinkResolver",
    "MultiFlowRouter",
    "PFloodSinkResolver",
    "SingleFlowRouter",
]



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))
import os
import subprocess

from matplotlib.font_manager import FontManager

# -- Doxygen build on RTD ----------------------------------------------------

on_rtd = os.environ.get("READTHEDOCS", None) == "True"

if on_rtd:
    subprocess.call("cd ..; doxygen", shell=True)

# -- Project information -----------------------------------------------------

project = "Fastscapelib"
copyright = "since 2018, Fastscapelib developers"
author = "Benoit Bovy"
# The short X.Y version
version = "0.2.1"
# The full version, including alpha/beta/rc tags
release = version

print("building doc for fastscapelib version ", version)

# -- General configuration ---------------------------------------------------

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.extlinks",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "breathe",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",
    "sphinx_remove_toctrees",
    "sphinxcontrib.bibtex",
    "sphinxcontrib.mermaid",
    "sphinx_design",
    "myst_nb",
]

extlinks = {
    "issue": ("https://github.com/fastscape-lem/fastscapelib/issues/%s", "#%s"),
    "pull": ("https://github.com/fastscape-lem/fastscapelib/pull/%s", "#%s"),
}

templates_path = ["_templates"]

source_suffix = [".rst", ".md"]

master_doc = "index"

language = "en"

exclude_patterns = [
    "**.ipynb_checkpoints",
    "build/**.ipynb",
    "Thumbs.db",
    ".DS_Store",
]

bibtex_bibfiles = ["references.bib"]

# pygments_style = "sphinx"
highlight_language = "c++"
myst_enable_extensions = ["colon_fence", "substitution"]

# not working yet (https://github.com/sphinx-doc/sphinx/issues/5379)
autodoc_member_order = "bysource"

napoleon_numpy_docstring = True
napoleon_use_param = True
napoleon_use_rtype = False
napoleon_preprocess_types = True
napoleon_type_aliases = {
    "array_like": ":term:`array_like`",
    "array-like": ":term:`array-like <array_like>`",
}
typehints_defaults = "comma"
typehints_use_rtype = False

remove_from_toctrees = ["api_python/_api_generated/*"]

# -- Myst-NB config ----------------------------------------------------------

# This is to mitigate errors on CI VMs, where you can get the message:
# Matplotlib is building the font cache" in output notebooks
FontManager()

nb_kernel_rgx_aliases = {".*fastscapelib.*": "python3"}

# -- Options for HTML output -------------------------------------------------

html_theme = "furo"
html_title = ""

html_theme_options = {
    # Furo Theme
    "sidebar_hide_name": True,
    "light_logo": "fastscapelib_logo.svg",
    "dark_logo": "fastscapelib_logo.svg",
    # "announcement": "Fastscapelib is in <em>active development</em> (refactoring) and will be released soon, stay tuned!",
    "footer_icons": [
        {
            "name": "GitHub",
            "url": "https://github.com/fastscape-lem/fastscapelib",
            "html": """
                <svg stroke="currentColor" fill="currentColor" stroke-width="0" viewBox="0 0 16 16">
                    <path fill-rule="evenodd" d="M8 0C3.58 0 0 3.58 0 8c0 3.54 2.29 6.53 5.47 7.59.4.07.55-.17.55-.38 0-.19-.01-.82-.01-1.49-2.01.37-2.53-.49-2.69-.94-.09-.23-.48-.94-.82-1.13-.28-.15-.68-.52-.01-.53.63-.01 1.08.58 1.23.82.72 1.21 1.87.87 2.33.66.07-.52.28-.87.51-1.07-1.78-.2-3.64-.89-3.64-3.95 0-.87.31-1.59.82-2.15-.08-.2-.36-1.02.08-2.12 0 0 .67-.21 2.2.82.64-.18 1.32-.27 2-.27.68 0 1.36.09 2 .27 1.53-1.04 2.2-.82 2.2-.82.44 1.1.16 1.92.08 2.12.51.56.82 1.27.82 2.15 0 3.07-1.87 3.75-3.65 3.95.29.25.54.73.54 1.48 0 1.07-.01 1.93-.01 2.2 0 .21.15.46.55.38A8.013 8.013 0 0 0 16 8c0-4.42-3.58-8-8-8z"></path>
                </svg>
            """,
            "class": "",
        },
    ],
}
html_static_path = ["_static"]
html_css_files = ["style.css"]
html_favicon = "_static/favicon.ico"
htmlhelp_basename = "fastscapelibdoc"

# -- Breathe configuration -------------------------------------------------

breathe_projects = {"fastscapelib": "../_xml"}
breathe_default_project = "fastscapelib"
breathe_show_include = True

# -- Options for intersphinx extension ---------------------------------------

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "xtensor": ("https://xtensor.readthedocs.io/en/latest/", None),
    "xtensor-python": ("https://xtensor-python.readthedocs.io/en/latest/", None),
}



from typing import overload

import numpy as np
import numpy.typing as npt

from fastscapelib.flow import FlowGraph
from fastscapelib.grid import RasterGrid

class SPLEroder:
    @overload
    def __init__(
        self,
        flow_graph: FlowGraph,
        k_coef: float,
        area_exp: float,
        slope_exp: float,
        tolerance: float,
    ) -> None: ...
    @overload
    def __init__(
        self,
        flow_graph: FlowGraph,
        k_coef: npt.NDArray[np.float64],
        area_exp: float,
        slope_exp: float,
        tolerance: float,
    ) -> None: ...
    @property
    def k_coef(self) -> npt.ArrayLike: ...
    @k_coef.setter
    def k_coef(self, value: npt.ArrayLike) -> None: ...
    @property
    def area_exp(self) -> float: ...
    @area_exp.setter
    def area_exp(self, value: float) -> None: ...
    @property
    def slope_exp(self) -> float: ...
    @slope_exp.setter
    def slope_exp(self, value: float) -> None: ...
    @property
    def tolerance(self) -> float: ...
    @property
    def n_corr(self) -> int: ...
    def erode(
        self,
        elevation: npt.NDArray[np.float64],
        drainage_area: npt.NDArray[np.float64],
        dt: float,
    ) -> npt.NDArray[np.float64]: ...

class DiffusionADIEroder:
    @overload
    def __init__(self, grid: RasterGrid, k_coef: float) -> None: ...
    @overload
    def __init__(self, grid: RasterGrid, k_coef: npt.NDArray[np.float64]) -> None: ...
    @property
    def k_coef(self) -> npt.ArrayLike: ...
    @k_coef.setter
    def k_coef(self, value: npt.ArrayLike) -> None: ...
    def erode(
        self, elevation: npt.NDArray[np.float64], dt: float
    ) -> npt.NDArray[np.float64]: ...



import numpy as np
import pytest

from fastscapelib.eroders import DiffusionADIEroder, SPLEroder
from fastscapelib.flow import FlowGraph, MultiFlowRouter, SingleFlowRouter
from fastscapelib.grid import NodeStatus, ProfileGrid, RasterGrid


class TestSPLEroder:
    def test_constructor_properties(self) -> None:
        grid = RasterGrid([2, 2], [1.0, 1.0], NodeStatus.FIXED_VALUE)
        flow_graph = FlowGraph(grid, [SingleFlowRouter()])

        eroder = SPLEroder(flow_graph, 1e-3, 0.4, 1, 1e-5)

        np.testing.assert_equal(eroder.k_coef, np.full(grid.size, 1e-3))
        assert eroder.area_exp == 0.4
        assert eroder.slope_exp == 1
        assert eroder.tolerance == 1e-5

        eroder.area_exp = 0.5
        assert eroder.area_exp == 0.5

        eroder.slope_exp = 1.2
        assert eroder.slope_exp == 1.2

        k_coef_arr = np.arange(grid.size, dtype=np.double).reshape(grid.shape)
        eroder.k_coef = k_coef_arr
        np.testing.assert_equal(eroder.k_coef, k_coef_arr.flatten())

        k_coef = 1e-5
        eroder.k_coef = k_coef
        np.testing.assert_equal(eroder.k_coef, np.full(grid.size, 1e-5))

        with pytest.raises(RuntimeError, match=".*shape mismatch"):
            eroder.k_coef = np.ones((4, 5))

        # multiple flow direction and n != 1 is not supoprted
        flow_graph2 = FlowGraph(grid, [MultiFlowRouter(1.0)])
        with pytest.raises(ValueError, match=".*slope exponent != 1 is not supported"):
            SPLEroder(flow_graph2, 1e-3, 0.4, 1.5, 1e-3)

    @pytest.mark.parametrize("k_coef", [1e-3, np.full((4), 1e-3)])
    def test_profile_grid(self, k_coef) -> None:
        spacing = 300.0
        grid = ProfileGrid(4, 300, NodeStatus.FIXED_VALUE)

        flow_graph = FlowGraph(grid, [SingleFlowRouter()])

        area_exp = 0.5
        slope_exp = 1.0
        tolerance = 1e-3
        eroder = SPLEroder(flow_graph, k_coef, area_exp, slope_exp, tolerance)

        h = 1.0
        elevation = np.array([0.0, h, h, 0.0], dtype="d")

        flow_graph.update_routes(elevation)
        drainage_area = flow_graph.accumulate(1.0)

        dt = 1.0  # use small time step (compare with explicit scheme)
        erosion = eroder.erode(elevation, drainage_area, dt)

        slope = h / spacing
        a = spacing
        k_coef = 1e-3
        err = dt * k_coef * a**area_exp * slope**slope_exp
        expected_erosion = np.array([0.0, err, err, 0.0], dtype="d")

        np.testing.assert_allclose(erosion, expected_erosion, atol=1e-5)
        assert eroder.n_corr == 0

    @pytest.mark.parametrize("k_coef", [1e-3, np.full((2, 2), 1e-3)])
    def test_raster_grid(self, k_coef) -> None:
        # Test on a tiny (2x2) 2-d square grid with a planar surface
        # tilted in y (rows) and with all outlets on the 1st row.
        spacing = 300.0

        # top-border base level
        top_base_level = [
            NodeStatus.CORE,
            NodeStatus.CORE,
            NodeStatus.FIXED_VALUE,
            NodeStatus.CORE,
        ]

        grid = RasterGrid([2, 2], [spacing, spacing], top_base_level)

        flow_graph = FlowGraph(grid, [SingleFlowRouter()])

        area_exp = 0.5
        slope_exp = 1.0
        tolerance = 1e-3
        eroder = SPLEroder(flow_graph, k_coef, area_exp, slope_exp, tolerance)

        h = 1.0
        elevation = np.array([[0.0, 0.0], [h, h]], dtype="d")

        flow_graph.update_routes(elevation)
        drainage_area = flow_graph.accumulate(1.0)

        dt = 1  # use small time step (compare with explicit scheme)
        erosion = eroder.erode(elevation, drainage_area, dt)

        slope = h / spacing
        a = spacing**2
        k_coef = 1e-3
        err = dt * k_coef * a**area_exp * slope**slope_exp
        expected_erosion = np.array([[0.0, 0.0], [err, err]], dtype="d")

        np.testing.assert_allclose(erosion, expected_erosion, atol=1e-5)
        assert eroder.n_corr == 0


def _solve_diffusion_analytical(x, y, k_coef, t):
    fact = 4 * k_coef * t
    return np.exp(-(x * x + y * y) / fact) / (fact * np.pi)


def _compute_l2_norm(a1, a2):
    return 1.0 / a1.size * np.sum(a2**2 - a1**2)


class TestDiffusionADIEroder:
    def test_constructor_properties(self) -> None:
        grid = RasterGrid([2, 2], [1.0, 1.0], NodeStatus.FIXED_VALUE)

        eroder = DiffusionADIEroder(grid, 1e-3)

        np.testing.assert_equal(eroder.k_coef, np.full(grid.shape, 1e-3))

        k_coef_arr = np.full(grid.shape, 1e-2)
        eroder.k_coef = k_coef_arr
        np.testing.assert_equal(eroder.k_coef, k_coef_arr)

        k_coef = 1e-5
        eroder.k_coef = k_coef
        np.testing.assert_equal(eroder.k_coef, np.full(grid.shape, 1e-5))

        with pytest.raises(RuntimeError, match=".*shape mismatch"):
            eroder.k_coef = np.ones((4, 5))

    @pytest.mark.parametrize("k_coef_type", ["scalar", "array"])
    def test_erode_linear_diffusion(self, k_coef_type) -> None:
        x, y = np.meshgrid(np.linspace(-20, 20, 51), np.linspace(-20, 20, 101))
        dy = 0.4
        dx = 0.8

        grid = RasterGrid([101, 51], [dy, dx], NodeStatus.FIXED_VALUE)

        k_coef = 1e-3
        eroder = DiffusionADIEroder(grid, k_coef)

        dt = 1e3
        t0 = 2e3

        elevation_init = _solve_diffusion_analytical(x, y, k_coef, t0)
        erosion = np.empty_like(elevation_init)

        elevation_analytical = _solve_diffusion_analytical(x, y, k_coef, t0 + dt)

        if k_coef_type == "array":
            eroder.k_coef = np.full_like(x, k_coef)

        erosion = eroder.erode(elevation_init, dt)
        elevation_numerical = elevation_init - erosion

        l2_norm = _compute_l2_norm(elevation_analytical, elevation_numerical)

        assert l2_norm < 1e-9
